package nl.kallestruik.dtweaks.fakeplayer

import com.destroystokyo.paper.event.player.PlayerInitialSpawnEvent
import com.mojang.authlib.GameProfile
import com.mojang.datafixers.util.Either
import com.mojang.serialization.Dynamic
import io.netty.buffer.Unpooled
import io.papermc.paper.configuration.GlobalConfiguration
import net.minecraft.core.RegistryAccess
import net.minecraft.nbt.CompoundTag
import net.minecraft.nbt.NbtOps
import net.minecraft.nbt.Tag
import net.minecraft.network.Connection
import net.minecraft.network.FriendlyByteBuf
import net.minecraft.network.protocol.game.*
import net.minecraft.resources.ResourceKey
import net.minecraft.server.MinecraftServer
import net.minecraft.server.level.ChunkHolder.ChunkLoadingFailure
import net.minecraft.server.level.DistanceManager
import net.minecraft.server.level.ServerLevel
import net.minecraft.server.level.ServerPlayer
import net.minecraft.server.level.TicketType
import net.minecraft.server.network.ServerGamePacketListenerImpl
import net.minecraft.server.players.GameProfileCache
import net.minecraft.server.players.PlayerList
import net.minecraft.tags.TagNetworkSerialization
import net.minecraft.world.level.ChunkPos
import net.minecraft.world.level.GameRules
import net.minecraft.world.level.Level
import net.minecraft.world.level.biome.BiomeManager
import net.minecraft.world.level.chunk.ChunkAccess
import net.minecraft.world.level.dimension.DimensionType
import nl.kallestruik.dlib.ReflectionUtil
import org.bukkit.Bukkit
import org.bukkit.craftbukkit.v1_19_R1.CraftServer
import org.bukkit.craftbukkit.v1_19_R1.CraftWorld
import org.bukkit.entity.Player
import org.spigotmc.event.player.PlayerSpawnLocationEvent
import java.util.*
import java.util.concurrent.CompletableFuture

object FakePlayerList {

    fun PlayerList.placeFakePlayer(connection: Connection, player: ServerPlayer) {
        player.isRealPlayer = true // Paper - Chunk priority
        player.networkManager = connection // Paper
        player.loginTime = System.currentTimeMillis() // Paper

        val gameprofile = player.getGameProfile()
        val usercache = server.profileCache
        val optional = usercache[gameprofile.id]
        var s: String? = optional.map { obj: GameProfile -> obj.name }.orElse(gameprofile.name)

        usercache.add(gameprofile)
        val nbttagcompound = load(player)
        // CraftBukkit start - Better rename detection
        // CraftBukkit start - Better rename detection
        if (nbttagcompound != null && nbttagcompound.contains("bukkit")) {
            val bukkit = nbttagcompound.getCompound("bukkit")
            s = if (bukkit.contains("lastKnownName", 8)) bukkit.getString("lastKnownName") else s
        }
        val lastKnownName = s // Paper

        // CraftBukkit end

        // Paper start - move logic in Entity to here, to use bukkit supplied world UUID.
        // CraftBukkit end

        // Paper start - move logic in Entity to here, to use bukkit supplied world UUID.
        val resourcekey =
            if (nbttagcompound != null && nbttagcompound.contains("WorldUUIDMost") && nbttagcompound.contains("WorldUUIDLeast")) {
                val uid = UUID(nbttagcompound.getLong("WorldUUIDMost"), nbttagcompound.getLong("WorldUUIDLeast"))
                val bWorld = Bukkit.getServer().getWorld(uid)
                if (bWorld != null) {
                    (bWorld as CraftWorld).handle.dimension()
                } else {
                    Level.OVERWORLD
                }
            } else if (nbttagcompound != null) {
                // Vanilla migration support
                // Paper end
                val dataresult = DimensionType.parseLegacy(
                    Dynamic(
                        NbtOps.INSTANCE,
                        nbttagcompound["Dimension"]
                    )
                ) // CraftBukkit - decompile error
                dataresult.resultOrPartial { msg: String? ->

                }.orElse(Level.OVERWORLD)
            } else {
                Level.OVERWORLD
            }

        val resourcekey1: ResourceKey<Level> = resourcekey
        val worldserver = server.getLevel(resourcekey1)
        var worldserver1: ServerLevel

        worldserver1 = if (worldserver == null) {
            server.overworld()
        } else {
            worldserver
        }

        if (nbttagcompound == null) player.fudgeSpawnLocation(worldserver1) // Paper - only move to spawn on first login, otherwise, stay where you are....


        player.setLevel(worldserver1)
        // Paper start - make s1 final
        // Paper start - make s1 final
        val s1: String

        s1 = if (connection.remoteAddress != null) {
            if (GlobalConfiguration.get().logging.logPlayerIpAddresses) connection.remoteAddress.toString() else "<ip address withheld>" // Paper
        } else {
            "local"
        }
        // Paper end

        // Spigot start - spawn location event
        // Paper end

        // Spigot start - spawn location event
        val spawnPlayer: Player = player.bukkitEntity
        val ev: PlayerSpawnLocationEvent =
            PlayerInitialSpawnEvent(spawnPlayer, spawnPlayer.location) // Paper use our duplicate event

        val cserver = ReflectionUtil().getValueFromField(this, "cserver") as CraftServer

        cserver.pluginManager.callEvent(ev)

        val loc = ev.spawnLocation
        worldserver1 = (loc.world as CraftWorld).handle

        player.spawnIn(worldserver1)
        player.gameMode.setLevel(player.level as ServerLevel)
        player.setPosRaw(loc.x, loc.y, loc.z)
        player.setRot(loc.yaw, loc.pitch)

        val worlddata = worldserver1.getLevelData()

        player.loadGameTypes(nbttagcompound)
        val playerconnection = ServerGamePacketListenerImpl(server, connection, player)
        val gamerules = worldserver1.gameRules
        val flag = gamerules.getBoolean(GameRules.RULE_DO_IMMEDIATE_RESPAWN)
        val flag1 = gamerules.getBoolean(GameRules.RULE_REDUCEDDEBUGINFO)

        val registryHolder = ReflectionUtil().getValueFromField(this, "t" /* registryHolder -> t */) as RegistryAccess.Frozen

        playerconnection.send(
            ClientboundLoginPacket(
                player.id,
                worlddata.isHardcore,
                player.gameMode.gameModeForPlayer,
                player.gameMode.previousGameModeForPlayer,
                server.levelKeys(),
                registryHolder,
                worldserver1.dimensionTypeId(),
                worldserver1.dimension(),
                BiomeManager.obfuscateSeed(worldserver1.seed),
                maxPlayers,
                worldserver1.getChunkSource().chunkMap.playerChunkManager.targetSendDistance,
                worldserver1.getChunkSource().chunkMap.playerChunkManager.targetTickViewDistance,
                flag1,
                !flag,
                worldserver1.isDebug,
                worldserver1.isFlat,
                player.lastDeathLocation
            )
        )

        player.bukkitEntity.sendSupportedChannels() // CraftBukkit

        playerconnection.send(
            ClientboundCustomPayloadPacket(
                ClientboundCustomPayloadPacket.BRAND, FriendlyByteBuf(Unpooled.buffer()).writeUtf(
                    server.serverModName
                )
            )
        )
        playerconnection.send(ClientboundChangeDifficultyPacket(worlddata.difficulty, worlddata.isDifficultyLocked))
        playerconnection.send(ClientboundPlayerAbilitiesPacket(player.abilities))
        playerconnection.send(ClientboundSetCarriedItemPacket(player.inventory.selected))
        playerconnection.send(ClientboundUpdateRecipesPacket(server.recipeManager.getRecipes()))
        playerconnection.send(ClientboundUpdateTagsPacket(TagNetworkSerialization.serializeTagsToNetwork(registryHolder)))
        this.sendPlayerPermissionLevel(player)
        player.stats.markAllDirty()
        player.recipeBook.sendInitialRecipeBook(player)
        updateEntireScoreboard(worldserver1.scoreboard, player)
        server.invalidateStatus()
        // Paper start - async load spawn in chunk
        // Paper start - async load spawn in chunk
        val finalWorldserver = worldserver1
        val chunkX = loc.blockX shr 4
        val chunkZ = loc.blockZ shr 4
        val pos = ChunkPos(chunkX, chunkZ)
        val playerChunkMap = worldserver1.getChunkSource().chunkMap
        val distanceManager: DistanceManager = playerChunkMap.distanceManager
        distanceManager.addTicket(TicketType.LOGIN, pos, 31, pos.toLong())
        worldserver1.getChunkSource().markAreaHighPriority(pos, 28, 3) // Paper - Chunk priority

        val postChunkLoadJoin = this::class.java.superclass.getDeclaredMethod("postChunkLoadJoin", ServerPlayer::class.java, ServerLevel::class.java, Connection::class.java, ServerGamePacketListenerImpl::class.java, CompoundTag::class.java, String::class.java, String::class.java)
        postChunkLoadJoin.isAccessible = true
        postChunkLoadJoin.invoke(this, player, finalWorldserver, connection, playerconnection,
            nbttagcompound, "127.0.0.1", lastKnownName)
    }

    fun placeNewPlayer(playerList: PlayerList, server: MinecraftServer, connection: FakeConnection, player: FakePlayer) {

    }

        fun placeNewPlayerOld(playerList: PlayerList, server: MinecraftServer, connection: FakeConnection, player: FakePlayer) {
        player.isRealPlayer = true // Paper - Chunk priority

        player.networkManager = connection // Paper

        player.loginTime = System.currentTimeMillis() // Paper

        val gameProfile: GameProfile = player.getGameProfile()
        val userCache: GameProfileCache = server.getProfileCache()
        val optional = userCache[gameProfile.id]
        var s: String? = optional.map { obj: GameProfile -> obj.name }.orElse(gameProfile.name)

        userCache.add(gameProfile)
        val nbtTagCompound: CompoundTag? = playerList.load(player)
        // CraftBukkit start - Better rename detection
        // CraftBukkit start - Better rename detection
        if (nbtTagCompound != null && nbtTagCompound.contains("bukkit")) {
            val bukkit = nbtTagCompound.getCompound("bukkit")
            s = if (bukkit.contains("lastKnownName", 8)) bukkit.getString("lastKnownName") else s
        }
        val lastKnownName = s // Paper

        // CraftBukkit end

        // Paper start - move logic in Entity to here, to use bukkit supplied world UUID.
        // CraftBukkit end

        // Paper start - move logic in Entity to here, to use bukkit supplied world UUID.
        val resourceKey: ResourceKey<Level> = if (nbtTagCompound != null && nbtTagCompound.contains("WorldUUIDMost") && nbtTagCompound.contains("WorldUUIDLeast")) {
                val uid = UUID(nbtTagCompound.getLong("WorldUUIDMost"), nbtTagCompound.getLong("WorldUUIDLeast"))
                val bWorld = Bukkit.getServer().getWorld(uid)
                if (bWorld != null) {
                    (bWorld as CraftWorld).handle.dimension()
                } else {
                    Level.OVERWORLD
                }
            } else if (nbtTagCompound != null) {
                // Vanilla migration support
                // Paper end
                val dataResult = DimensionType.parseLegacy(
                    Dynamic(
                        NbtOps.INSTANCE,
                        nbtTagCompound["Dimension"]
                    )
                )
                dataResult.result().orElse(Level.OVERWORLD)
            } else {
                Level.OVERWORLD
            }

        var worldServer: ServerLevel = server.getLevel(resourceKey) ?: server.overworld()

        if (nbtTagCompound == null) player.fudgeSpawnLocation(worldServer) // Paper - only move to spawn on first login, otherwise, stay where you are....

        player.setLevel(worldServer)


        // Spigot start - spawn location event
        val spawnPlayer: Player = player.bukkitEntity
        val ev: PlayerSpawnLocationEvent =
            PlayerInitialSpawnEvent(spawnPlayer, spawnPlayer.location) // Paper use our duplicate event

        Bukkit.getServer().pluginManager.callEvent(ev)

        val loc = ev.spawnLocation
        worldServer = (loc.world as CraftWorld).handle

        player.spawnIn(worldServer)
        player.gameMode.setLevel(player.level as ServerLevel)
        // Paper start - set raw so we aren't fully joined to the world (not added to chunk or world)
        // Paper start - set raw so we aren't fully joined to the world (not added to chunk or world)
        player.setPosRaw(loc.x, loc.y, loc.z)
        player.setRot(loc.yaw, loc.pitch)
        // Paper end
        // Spigot end

        // CraftBukkit - Moved message to after join
        // PlayerList.LOGGER.info("{}[{}] logged in with entity id {} at ({}, {}, {})", entityplayer.getName().getString(), s1, entityplayer.getId(), entityplayer.getX(), entityplayer.getY(), entityplayer.getZ());
        // Paper end
        // Spigot end

        // CraftBukkit - Moved message to after join
        // PlayerList.LOGGER.info("{}[{}] logged in with entity id {} at ({}, {}, {})", entityplayer.getName().getString(), s1, entityplayer.getId(), entityplayer.getX(), entityplayer.getY(), entityplayer.getZ());
        val worlddata = worldServer.getLevelData()

        player.loadGameTypes(nbtTagCompound)
        val playerconnection = FakePlayerConnection(server, connection, player)
        val gamerules = worldServer.gameRules

        player.bukkitEntity.sendSupportedChannels() // CraftBukkit

        player.stats.markAllDirty()
        player.recipeBook.sendInitialRecipeBook(player)
        playerList.updateEntireScoreboard(worldServer.scoreboard, player)
        server.invalidateStatus()
        // Paper start - async load spawn in chunk
        // Paper start - async load spawn in chunk
        val finalWorldserver = worldServer
        val chunkX = loc.blockX shr 4
        val chunkZ = loc.blockZ shr 4
        val pos = ChunkPos(chunkX, chunkZ)
        val playerChunkMap = worldServer.getChunkSource().chunkMap
        val distanceManager: DistanceManager = playerChunkMap.distanceManager
//        distanceManager.addTicket(TicketType.LOGIN, pos, 31, pos.toLong())
        worldServer.getChunkSource().markAreaHighPriority(pos, 28, 3) // Paper - Chunk priority

        val postChunkLoadJoin = playerList::class.java.superclass.getDeclaredMethod("postChunkLoadJoin", ServerPlayer::class.java, ServerLevel::class.java, Connection::class.java, ServerGamePacketListenerImpl::class.java, CompoundTag::class.java, String::class.java, String::class.java)
        postChunkLoadJoin.isAccessible = true
        postChunkLoadJoin.invoke(playerList, player, finalWorldserver, connection, playerconnection,
                nbtTagCompound, "127.0.0.1", lastKnownName)


//        worldServer.getChunkSource().getChunkAtAsynchronously(chunkX, chunkZ, true, false)
//            .thenApply { chunk: Either<ChunkAccess, ChunkLoadingFailure> ->  // Paper - Chunk priority
//                val updatingChunk = playerChunkMap.getUpdatingChunkIfPresent(pos.toLong())
//                if (updatingChunk != null) {
//                    return@thenApply updatingChunk.entityTickingChunkFuture
//                } else {
//                    return@thenApply CompletableFuture.completedFuture(
//                        chunk
//                    )
//                }
//            }.thenAccept {
//
//        }
    }
}